/** @file massiv_rsa.h
 * RSA cipher.
 */

/*
 * begin        : Tue Dec 18 2001
 * original     : Wei Dai (file version)
 * modified     : (C) 2001 by Ondrej Pecta (string version), Marek Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_MASSIV_RSA_H
#define MASSIV_MASSIV_RSA_H

#ifndef MASSIV_COMMON_H
#include "global/common.h"
#endif

#ifndef MASSIV_TYPES_H
#include "utility/types.h"
#endif

// CryptoPP

#include "net/cryptopp_begin.h"
#include "cryptopp/rsa.h"
#include "cryptopp/randpool.h"
#include "cryptopp/filters.h"
#include "net/cryptopp_end.h"

#include <vector>


namespace Massiv {
namespace Core {


/** @addtogroup net
 * @{
 */

/**
 * RSACipher.
 * @todo Describe the class properly.
 */
class RSACipher
    {
public:

    typedef CryptoPP::RSAES_OAEP_SHA_Decryptor      Decryptor;
    typedef CryptoPP::RSAES_OAEP_SHA_Encryptor      Encryptor;
    typedef CryptoPP::RSASSA_PKCS1v15_SHA_Signer    Signer;
    typedef CryptoPP::RSASSA_PKCS1v15_SHA_Verifier  Verifier;

    enum
        {
        MAX_BUFFER_LENGTH = 128
        };

public:

    struct Keys
        {
        friend class RSACipher;

    private:

        std::string public_key;
            /**< Public key. */
        std::string private_key;
            /**< Private key. */

        }; // Keys

    typedef std::string Seed;

public:

    /**
     * Default constructor.
     */
    RSACipher();

    /**
     * Initializes cipher and generates keys.
     */
    RSACipher
        (
        UInt32          key_length,
        const Seed &    seed
        );

    /**
     * Initializes cipher using given pair of keys.
     */
    RSACipher
        (
        const Keys &    keys,
        const Seed &    seed
        );

    /**
     * Destructor.
     */
    ~RSACipher();

private:

    /**
     * Initializes random pool.
     */
    void init_random_pool
        (
        const Seed &    seed
        );

public:

    //
    // Single-buffer oriented operations.
    //

    /**
     * Encrypts @a message of length @a message_length
     * into to a temporary buffer.
     * Returns length of encrypted message.
     * Uses public key.
     */
    UInt32 encrypt
        (
        const UInt8 *   message,
        UInt32          message_length
        );

    /**
     * Encrypts @a message of length @a message_length
     * into to a temporary buffer.
     * Returns length of encrypted message.
     * Uses @a other_public_key public key.
     */
    UInt32 encrypt_other_public
        (
        const UInt8 *       message,
        UInt32              message_length,
        const std::string & other_public_key
        );

    /**
     * Decrypts @a encrypted into a temporary buffer.
     * Returns length of decrypted message.
     * Uses private key.
     */
    UInt32 decrypt
        (
        const UInt8 *   encrypted,
        UInt32          encrypted_length
        );

    /**
     * Signs @a message of lenght @a message_length
     * and places the signature into a temporary buffer.
     * Returns length of the signature.
     */
    UInt32 sign
        (
        const UInt8 *   message,
        UInt32          message_length
        );

    /**
     * Verifies the message (i.e. checks whether the @a signature
     * matches the @a message).
     */
    bool verify
        (
        const UInt8 *   message,
        UInt32          message_length,
        const UInt8 *   signature,
        UInt32          signature_length
        );

    /**
     * Verifies the message (i.e. checks whether the @a signature
     * matches the @a message) with other public key.
     */
    bool verify_other_public
        (
        const UInt8 *       message,
        UInt32              message_length,
        const UInt8 *       signature,
        UInt32              signature_length,
        const std::string & other_public_key
        );

    /**
     * Returns content of the temporary buffer.
     */
    const UInt8 * get_output_buffer() const;

    /**
     * Returns max size of plain text which can be encoded with this 
     * RSA ciphre. Depends on RSA ciphre length.
     */
    UInt32  get_max_text_len_for_encoding
        (
        const std::string & other_public_key
        ) const;

public:

    //
    // Wrappers.
    //

    /**
     * Encrypts <first, last] range and places it to output.
     */
    template <class InputIterator, class OutputContainer>
    void encrypt_range
        (
        InputIterator           first,
        InputIterator           last,
        OutputContainer &       output
        )
        {
        int size;
        size = encrypt( copy_input( first, last, 0 ), last - first );
        copy_output( output, size );
        }

    /**
     * Encrypts <first, last] range with other public key and places it to output.
     */
    template <class InputIterator, class OutputContainer>
    void encrypt_range_other_public
        (
        InputIterator           first,
        InputIterator           last,
        OutputContainer &       output,
        std::string &           other_public
        )
        {
        int size;
        size = encrypt_other_public( copy_input( first, last, 0 ), last - first, other_public );
        copy_output( output, size );
        }

    /**
     * Encrypts message into encrypted.
     */
    template <class InputContainer, class OutputContainer>
    void encrypt
        (
        const InputContainer &  message,
        OutputContainer &       encrypted
        )
        {
        encrypt_range( message.begin(), message.end(), encrypted );
        }

    /**
     * Encrypts message with other public key into encrypted.
     */
    template <class InputContainer, class OutputContainer>
    void encrypt
        (
        const InputContainer &  message,
        OutputContainer &       encrypted,
        std::string &           other_public
        )
        {
        encrypt_range_other_public( message.begin(), message.end(), encrypted, other_public );
        }

    /**
     * Decrypts encrypted message and places it to decrypted.
     */
    template <class InputContainer, class OutputContainer>
    UInt32 decrypt
        (
        const InputContainer &  encrypted,
        OutputContainer &       decrypted
        )
        {
        const UInt8 *   input;
        UInt32          size;

        input = copy_input( encrypted.begin(), encrypted.end(), 0 );
        size = decrypt( input, encrypted.size() );
        copy_output( decrypted, size );
        return size;
        }

    /**
     * Signs the message.
     */
    template <class InputContainer, class OutputContainer>
    void sign
        (
        const InputContainer &  message,
        OutputContainer &       signature
        )
        {
        const UInt8 *   input;
        int             size;

        input = copy_input( message.begin(), message.end(), 0 );
        size = sign( input, message.size() );
        copy_output( signature, size );
        }

    /**
     * Verifies the message.
     */
    template <class MessageContainer, class SignatureContainer>
    bool verify
        (
        const MessageContainer &    message,
        const SignatureContainer &  signature
        )
        {
        return verify
            (
            copy_input( message.begin(), message.end(), 0 ),
            message.size(),
            copy_input( signature.begin(), signature.end(), 1 ),
            signature.size()
            );
        }

    /**
     * Verifies the message with other key.
     */
    template <class MessageContainer, class SignatureContainer>
    bool verify
        (
        const MessageContainer &    message,
        const SignatureContainer &  signature,
        std::string &               other_public
        )
        {
        return verify_other_public
            (
            copy_input( message.begin(), message.end(), 0 ),
            message.size(),
            copy_input( signature.begin(), signature.end(), 1 ),
            signature.size(),
            other_public
            );
        }

private:

    /**
     * Copies data from output buffer to the output container.
     */
    template <class OutputContainer>
    void copy_output
        (
        OutputContainer &   container,
        UInt32              size
        )
        {
        typename OutputContainer::iterator it;

        container.resize( size );

        for( it = container.begin(); it != container.end(); it++ )
            {
            *it = ( typename OutputContainer::value_type ) output_buffer[ it - container.begin() ];
            }
        }

    /**
     * Copies data from input container to the input buffer.
     */
    template <class InputIterator>
    const UInt8 * copy_input
        (
        InputIterator   first,
        InputIterator   last,
        int             input_number
        )
        {
        UInt8 * buffer;
        int     i;

        hope( last - first <= MAX_BUFFER_LENGTH );

        i = 0;
        buffer = input_buffer[ input_number ];

        while( first < last )
            {
            buffer[ i ] = ( UInt8 ) *first;

            first++;
            i++;
            }

        return buffer;
        }

public:

    /**
     * Returns current keys.
     */
    const Keys & get_keys() const;

    /**
    * Returns current public key.
    */
    const std::string & get_public_key() const;

    /**
    * Returns current private key.
    */
    const std::string & get_private_key() const;

    /**
     * Operator = .
     */
    RSACipher & operator=
        (
        const RSACipher & rsa
        );

    /**
     * Sets keys.
     */
    void set_keys
        (
        const std::string & private_key,
        const std::string & public_key
        );

private:

    Keys                    keys;
        /**< Current pair of keys. */
    CryptoPP::RandomPool    random_pool;
        /**< Random pool. */
    UInt8                   output_buffer[ MAX_BUFFER_LENGTH ];
        /**< Temporary buffer used for storing resuts of cipher operations.
             Note that those operations that use output_buffer can not
             be implemented in other way (ie. writing directly to
             output container) because we do not know how long will be the
             resulting message until the operation is performed. */
    UInt8                   input_buffer[ 2 ][ MAX_BUFFER_LENGTH ];
        /**< Two input buffers. */

    }; // RSACipher

/** @} */ // addtogroup net


} // namespace Massiv::Core
} // namespace Massiv

#endif // MASSIV_MASSIV_RSA_H
