/** @file debug.cpp
 * Debugging stuff.
 */

/*
 * begin        : Sat Oct 12 2002
 * copyright    : (C) 2002 by Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#include "utility/debug.h"
#include "global/global.h"
#include "thread/stack_tracer.h"
#include <cstdlib>
#include <iostream>
#include <sstream>


namespace Massiv {
namespace Core {


namespace {


/**
 * Why has program died?
 */
enum DieReason
    {
    DIE_ASSERT,
    DIE_PANIC
    }; // enum DieReason


/**
 * Print assert-like message, dump core and exit.
 * Never returns.
 */
NORETURN( void ) debug_die
    (
    const DieReason     reason,
    const char * const  file,
    const unsigned      line,
    const char * const  function,
    const char * const  message
    ) throw()
    {
    std::ostringstream ss;

    // Perform potentially dangerous actions at most once.

    static bool dying = false;
    bool recursed = false;
    if( dying )
        {
        recursed = true;
        }
    else
        {
        dying = true;
        }

    // Format message.

    switch( reason )
        {
        case DIE_ASSERT:
            {
            ss << "Assertion failed: ";
            break;
            }
        case DIE_PANIC:
            {
            ss << "Panic: ";
            break;
            }
        default:
            {
            ss << "Bad die argument. ";
            break;
            }
        }
    ss << message << " (file '" << file << "', line " << line;
    if( function )
        {
        ss << ", function " << function;
        }
    ss << ")";

    // Dump stack backtrace.

    if( !recursed && Global::stack_tracer_available() )
        {
        ss << "\nStack backtrace:\n";
        Global::stack_tracer().dump_stack( ss );
        }
    else
        {
        ss << "\nStack backtrace not available.";
        }

    // Send message to logger or std::cerr.
    
    const std::string complete_message = ss.str();

    bool logged = false;
    if( !recursed )
        {
        if( Global::logger_available() )
            {
            if( Global::log_error( Status::FACILITY_GENERIC, complete_message  ) )
                {
                logged = true;
                }
            }
        }

    if( !logged )
        {
        // Error message has not been logged, so print it to stderr.

        std::cerr << complete_message << std::endl;
        }
    else
        {
        // Error message has been logged. Maybe even to stderr.

        std::cerr << "Fatal error. Consult logs for more information." << std::endl;
        }
    
    #if defined( COMPILER_MSVC ) && defined( OPTION_DEBUG )

    // Ignore reason and use internal assert function under MSVC as it shows
    // nice messages, informs debugger and stuff...

    _assert( const_cast< char * >( message ), const_cast< char * >( file ), line );

    #else // COMPILER_MSVC

    // Use system/compiler-specific assert handler on DEBUG.

    #ifdef OPTION_DEBUG
    
    if( reason == DIE_ASSERT )
        {
        #if defined( COMPILER_GCC )

        __assert_fail( message, file, line, function );

        #endif // compiler
        }

    #endif

    // Abort.

    std::abort();
    std::exit( EXIT_FAILURE );

    #endif // COMPILER_MSVC
    }


} // anonymous namespace


NORETURN( void ) MASSIV_EXPORTED debug_panic
    (
    const char * const  file,
    const unsigned      line,
    const char * const  function,
    const char * const  message
    ) throw()
    {
    debug_die( DIE_PANIC, file, line, function, message );
    }


NORETURN( void ) MASSIV_EXPORTED debug_assert
    (
    const char * const  file,
    const unsigned      line,
    const char * const  function,
    const char * const  message
    ) throw()
    {
    debug_die( DIE_ASSERT, file, line, function, message );
    }


} // namespace Massiv::Core
} // namespace Massiv
